/*  Copyright 2009 Marc Toussaint
    email: mtoussai@cs.tu-berlin.de

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a COPYING file of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/> */

#ifndef MT_featherstone_h
#define MT_featherstone_h

#include "ors.h"

/*! interface and implementation to Featherstone's Articulated Body Algorithm

  See resources from http://users.rsise.anu.edu.au/~roy/spatial/index.html

  This is a direct port of the following two files
  http://users.rsise.anu.edu.au/~roy/spatial/ws04spatial.txt
  http://users.rsise.anu.edu.au/~roy/spatial/ws04abadyn.txt

  See also his slides on the spatial vector algebra
  http://users.rsise.anu.edu.au/~roy/spatial/slidesX4.pdf

  main changes for porting to C:

  indexing of arrays start from 0

  referencing sub arrays with [i] rather than {i}

  NOTE: Featherstone's rotation matricies have opposite convention than mine

  \ingroup sl
*/
namespace Featherstone{
  
  /*! Abstraction of the ors::Graph data to interface Featherstone's routines \ingroup sl */
  /*struct Robot{
    ors::Graph *C;
    uint N() const;
    int parent(uint i) const;
    byte dof(uint i) const;
    const arr S(uint i) const;
    const arr Xlink(uint i) const;
    const arr Ilink(uint i) const;
    const arr force(uint i) const;
  };*/

  struct Link{
    int type;
    int index;
    int parent;
    ors::Frame X,A,Q;
    ors::Vector com,force,torque;
    double mass;
    ors::Matrix inertia;
    uint dof(){ if(type==JHINGE) return 1; else return 0; }

    arr _h,_A,_Q,_I,_f; //featherstone types
    void setFeatherstones();
    void updateFeatherstones();
  };

  typedef MT::Array<Link> LinkTree;
  
  //! returns a cross-product matrix X such that \f$v \times y = X y\f$
  void skew(arr& X,const double *v);
  
  //! as above
  arr skew(const double *v);
  
  /*!\brief MM6 coordinate transform from X-axis rotation.  Xrotx(h)
    calculates the MM6 coordinate transform matrix (for motion
    vectors) induced by a rotation about the +X axis by an angle h (in
    radians).  Positive rotation is anticlockwise: +Y axis rotates
    towards +Z axis.
  */
  void Xrotx(arr& X, double h);
  
  /*!\brief MM6 coordinate transform from Y-axis rotation.  Xroty(h)
    calculates the MM6 coordinate transform matrix (for motion
    vectors) induced by a rotation about the +Y axis by an angle h (in
    radians).  Positive rotation is anticlockwise: +Z axis rotates
    towards +X axis.
  */
  void Xroty(arr& X, double h);
  
  /*!\brief MM6 coordinate transform from Z-axis rotation.  Xrotz(h)
    calculates the MM6 coordinate transform matrix (for motion
    vectors) induced by a rotation about the +Z axis by an angle h (in
    radians).  Positive rotation is anticlockwise: +X axis rotates
    towards +Y axis.
  */
  void Xrotz(arr& X, double h);
  
  
  /*!\brief MM6 coordinate transform from 3D translation vector.
    Xtrans(r) calculates the MM6 coordinate transform matrix (for
    motion vectors) induced by a shift of origin specified by the 3D
    vector r, which contains the x, y and z coordinates of the new
    location of the origin relative to the old.
  */
  void Xtrans(arr& X, double* r);
  
  
  /*!\brief Calculate RBI from mass, CoM and rotational inertia.
    RBmci(m,c,I) calculate MF6 rigid-body inertia tensor for a body
    with mass m, centre of mass at c, and (3x3) rotational inertia
    about CoM of I.
  */
  void RBmci(arr& rbi, double m, double *c, const ors::Matrix& I);
  
  /*!\brief MM6 cross-product tensor from M6 vector.  crossM(v)
    calculates the MM6 cross-product tensor of motion vector v such
    that crossM(v) * m = v X m (cross-product of v and m) where m is
    any motion vector or any matrix or tensor mapping to M6.
  */
  void crossM(arr& vcross,const arr& v);
  
  //! as above
  arr crossM(const arr& v);
  
  /*!\brief FF6 cross-product tensor from M6 vector.  crossF(v)
    calculates the FF6 cross-product tensor of motion vector v such
    that crossF(v) * f = v X f (cross-product of v and f) where f is
    any force vector or any matrix or tensor mapping to F6.
  */
  void crossF(arr& vcross,const arr& v);
  
  //! as above
  arr crossF(const arr& v);
  
  //void invdyn_old(arr& tau,const Robot& robot,const arr& qd,const arr& qdd,const arr& grav);
  
  void equationOfMotion(arr& M, arr& F, const LinkTree& tree,  const arr& qd);

  void fwdDynamics_MF    (arr& qdd, const LinkTree& tree, const arr& qd, const arr& tau);
  void fwdDynamics_aba_nD(arr& qdd, const LinkTree& tree, const arr& qd, const arr& tau);
  void fwdDynamics_aba_1D(arr& qdd, const LinkTree& tree, const arr& qd, const arr& tau);
  //void fwdDynamics_old(arr& qdd, const Robot& robot,   const arr& qd, const arr& tau, const arr& grav);

  void invDynamics       (arr& tau, const LinkTree& tree, const arr& qd, const arr& qdd);

}

void GraphToTree(MT::Array<Featherstone::Link>& tree, const ors::Graph& C);
void updateGraphToTree(MT::Array<Featherstone::Link>& tree, const ors::Graph& C);

#ifdef MT_IMPLEMENTATION
#include "ors_featherstone.cpp"
#endif

#endif
