/*  Copyright 2009 Marc Toussaint
    email: mtoussai@cs.tu-berlin.de

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a COPYING file of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/> */

#include "soc.h"
#include "opengl.h"
#include "plot.h"

//===========================================================================
//
// general documentation
//

/*! \brief libSOC -- Stochastic Optimal Control library

    This is the core namespace of libSOC.  See the <a
    href="../guide.pdf">guide</a> for an introduction.

    Please see also the header <a href="soc_8h-source.html">MT/soc.h</a> . */
namespace soc{};


//===========================================================================
//
// trivial helpers
//

//! \brief get the velocity vt of a trajectory q at time t
void soc::getVelocity(arr& vt,const arr& q,uint t,double tau){
  if(!t) vt = (q[0]-q[0])/tau;
  else   vt = (q[t]-q[t-1])/tau;
}

  //! compute the full (q,v) trajectory from a trajectory q
void soc::getPhaseTrajectory(arr& _q,const arr& q,double tau){
  uint T=q.d0,n=q.d1,t;
  arr vt;
  _q.resize(T,2,n);
  for(t=0;t<T;t++){
    getVelocity(vt,q,t,tau);
    _q.subDim(t,0)=q[t];
    if(t) _q.subDim(t,1)=(q[t]-q[t-1])/tau;
    else  _q.subDim(t,1)=0.;
  }
  _q.reshape(T,2*n);
}

//! simply get the q-trajectory from a (q,v)-trajectory
void soc::getPositionTrajectory(arr& q,const arr& _q){
  uint T=_q.d0,n=_q.d1/2,i,t;
  CHECK(2*n==_q.d1,"")
  q.resize(T,n);
  for(t=0;t<T;t++) for(i=0;i<n;i++) q(t,i)=_q(t,i);
}

/*! \brief use regularized Inverse Kinematics to compute a joint
    trajectory from a given task trajectory x for the 0-th task variable */
void soc::getJointFromTaskTrajectory(SocAbstraction& soci,arr& q,const arr& x){
  uint t,T=x.d0,n=soci.qDim();
  arr phiq,J,tJ,invJ,W,invW;
  q.resize(T,n);
  soci.getq0(q[0]());
  soci.getW(W);
  inverse_SymPosDef(invW,W);
  for(t=1;t<T;t++){
    soci.setq(q[t-1]);
    soci.getPhi(phiq,0);
    soci.getJtJ(J,tJ,0);
    pseudoInverse(invJ,J,invW,1e-5);
    q[t]() = q[t-1] + invJ*(x[t]-phiq);
  }
}

/*! \brief use regularized Inverse Kinematics to compute a joint
    trajectory from the task trajectory previously specifies for the
    taskid-th task variable */
void soc::straightTaskTrajectory(SocAbstraction& soci,arr& q,uint taskid){
  uint t,T=soci.nTime(),n=soci.qDim();
  arr phiq,xt,J,tJ,invJ,W,invW;
  q.resize(T,n);
  soci.getq0(q[0]());
  soci.getW(W);
  inverse_SymPosDef(invW,W);
  for(t=1;t<T;t++){
    soci.setq(q[t-1]);
    soci.getPhi(phiq,taskid);
    soci.getJtJ(J,tJ,taskid);
    soci.getTarget(xt,taskid,t);
    pseudoInverse(invJ,J,invW,1e-5);
    q[t]() = q[t-1] + invJ*(xt-phiq);
  }
  
}

//! not-yet-implemented
void soc::partialJointFromTaskTrajectory(SocAbstraction& soci,arr& dCdx,const arr& delCdelq,const arr& q,const arr& x){
  NIY;
}

//===========================================================================
//
// SocAbstraction defines routines that optimizers want access to
//

soc::SocAbstraction::SocAbstraction(){
  gl=NULL;
  os=NULL;
}

soc::SocAbstraction::~SocAbstraction(){
  //if(gl) delete gl;
}

void soc::SocAbstraction::getHessian(arr& H_i,uint i){ NIY; }

void soc::SocAbstraction::getProcess(arr& A,arr& a,arr& B){
  uint n=qDim();
  if(!dynamic){
    A.setId(n);
    B.setId(n);
    a.resize(n);
    a.setZero();
  }else{
    double tau=getTau();
    arr I,Z,Minv,F;
    I.setId(n);
    Z.resize(n,n); Z.setZero();
    
    getMinvF(Minv,F);
    
    A.setBlockMatrix(I,tau*I,Z,I);
    //double alpha = .1; //with fricion
    //A.setBlockMatrix(I,tau*I-tau*alpha*Minv,Z,I-tau*alpha*Minv);
    
    B.resize(2*n,n);
    B.setZero();
    B.setMatrixBlock(tau*tau*Minv,0,0);
    B.setMatrixBlock(tau*Minv,n,0);

    a.resize(2*n);
    a.setZero();
    a.setVectorBlock(tau*tau*Minv*F,0);
    a.setVectorBlock(tau*Minv*F,n);
  }
}

void soc::SocAbstraction::getProcess(arr& A,arr& tA,arr& invA,arr& invtA,arr& a,arr& B,arr& tB){
  CHECK(dynamic,"only implemented for dynamic yet");
  uint n=qDim();
  double tau=getTau();
  arr I,Z,Minv,F;
  I.setId(n);
  Z.resize(n,n); Z.setZero();

  getMinvF(Minv,F);

  A    .setBlockMatrix(I,tau*I,Z,I);
  tA   .setBlockMatrix(I,Z,tau*I,I);
  invA .setBlockMatrix(I,(-tau)*I,Z,I);
  invtA.setBlockMatrix(I,Z,(-tau)*I,I);

  B.resize(2*n,n);
  B.setZero();
  B.setMatrixBlock(tau*tau*Minv,0,0);
  B.setMatrixBlock(tau*Minv,n,0);

  tB.resize(n,2*n);
  tB.setZero();
  tB.setMatrixBlock(tau*tau*Minv,0,0);
  tB.setMatrixBlock(tau*Minv,0,n);

  a.resize(2*n);
  a.setZero();
  a.setVectorBlock(tau*tau*Minv*F,0);
  a.setVectorBlock(tau*Minv*F,n);
}

double soc::SocAbstraction::getCosts(arr& R,arr& r,uint t,const arr& qt){
  uint i,m=nTasks(),n=qDim();
  if(!dynamic){ //kinematic
    arr phiHatQ,J,tJ,x;
    double C=0.;
    double prec;
    R.resize(n,n); R.setZero();
    r.resize(n);   r.setZero();
    for(i=0;i<m;i++) if(isConditioned(i,t)){
      getPhi      (phiHatQ,i);
      getTarget   (x,i,t);
      getPrecision(prec,i,t);
      C += prec*sqrDistance(x,phiHatQ);
      getJtJ      (J,tJ,i);
      R += prec*tJ*J;
      r -= 2.*prec*tJ*(x - phiHatQ + J*qt);
    }
    return C;
  }else{
    uint n2=2*n;
    arr phiHatQ,Jqd,J,tJ,x,v,Ri,ri;
    double C=0.;
    double prec,precv;
    R.resize(n2,n2); R.setZero();
    r.resize(n2);    r.setZero();
    Ri.resize(n2,n2); Ri.setZero();
    ri.resize(n2);    ri.setZero();
    for(i=0;i<m;i++) if(isConditioned(i,t)){
      getPhi       (phiHatQ,i);
      getTarget    (x,i,t);
      getPrecision (prec,i,t);
      C += prec*sqrDistance(x,phiHatQ);
      getJqd       (Jqd,i);
      getTargetV   (v,i,t);
      getPrecisionV(precv,i,t);
      C += precv*sqrDistance(v,Jqd);
      getJtJ       (J,tJ,i);
      Ri.setMatrixBlock(prec*tJ*J,0,0);
      Ri.setMatrixBlock(precv*tJ*J,n,n);
      ri.setVectorBlock(prec*tJ*(x - phiHatQ + J*qt),0);
      ri.setVectorBlock(precv*tJ*v,n);
      R += Ri;
      r -= 2.*ri;
    }
    return C;
  }
  HALT("shoudn't be here");
  return 0.;
}

void soc::SocAbstraction::constantTrajectory(arr& q){
  CHECK(!dynamic,"");
  uint t,T=nTime(),n=qDim();
  q.resize(T,n);
  getq0(q[0]());
  for(t=1;t<T;t++) q[t]() = q[0];
}

void soc::SocAbstraction::passiveDynamicTrajectory(arr& q){
  uint t,T=nTime(),n=qDim();
  double tau=getTau();
  arr qd,Minv,F;
  qd.resize(T,n);
  q .resize(T,n);
  getqv0(q[0](),qd[0]());
  for(t=0;t<T-1;t++){
    setqv(q[t],qd[t]);
    getMinvF(Minv,F);
    qd[t+1]() = qd[t] + tau*Minv*F;
    q [t+1]() = q [t] + tau*qd[t+1];
  }
}

void soc::SocAbstraction::controlledDynamicTrajectory(arr& q,const arr& u){
  uint t,T=nTime(),n=qDim();
  if(!dynamic){
    q.resize(T,n);
    getq0(q[0]());
    for(t=1;t<T;t++) q[t] = q[t-1]+u[t-1];
    return;
  }
  double tau=getTau();
  arr qd,Minv,F;
  qd.resize(T,n);
  q .resize(T,n);
  getqv0(q[0](),qd[0]());
  for(t=0;t<T-1;t++){
    setqv(q[t],qd[t]);
    getMinvF(Minv,F);
    qd[t+1]() = qd[t] + tau*Minv*(F+u[t]);
    q [t+1]() = q [t] + tau*qd[t+1];  //note: this is qd[t+1] on the RHS
#if 0
    arr xx,x,A,a,B;
    x.setBlockVector(q[t],qd[t]);
    getProcessDynamic(A,a,B);
    xx = A*x + a + B*u[t];
    cout <<"xx = " <<xx <<"\nq=" <<q[t+1] <<qd[t+1] <<endl;
#endif
  }
}

void soc::SocAbstraction::getControlFromTrajectory(arr& u,const arr& q){
  uint t,T=nTime(),n=qDim();
  if(!dynamic){
    u.resize(T,n);
    u[T-1]().setZero();
    for(t=0;t<T-1;t++) u[t] = q[t+1]-q[t];
    return;
  }
  double tau,tau_1,tau_2;
  tau=getTau();
  tau_1 = 1./tau;
  tau_2 = tau_1*tau_1;
  arr M,F;
  u.resize(T,n);
  u[T-1]().setZero();
  for(t=0;t<T-1;t++){
    if(!t) setq(q[t]);
    else   setqv(q[t],tau_1*(q[t]-q[t-1]));
    getMF(M,F);
    if(t<T-1 && t>0)
      u[t]() = tau_2*M*(q[t+1]+q[t-1]-2.*q[t]) - F;
    if(t==0)
      u[t]() = tau_2*M*(q[t+1]-q[t]) - F;
  }
}

double soc::getTaskCost(soc::SocAbstraction& soci,int t){
  uint m=soci.nTasks();
  uint i;
  double C=0.;
  arr phiHatQ,Jqd,x,v;
  double prec;
  for(i=0;i<m;i++) if(soci.isConditioned(i,t)){
    soci.getPhi      (phiHatQ,i);
    soci.getTarget   (x,i,t);
    soci.getPrecision(prec,i,t);
    C += prec*sqrDistance(x,phiHatQ);
    if(soci.dynamic){
      soci.getJqd       (Jqd,i);
      soci.getTargetV   (v,i,t);
      soci.getPrecisionV(prec,i,t);
      C += prec*sqrDistance(v,Jqd);
    }
  }
  return C;
}

void soc::getTaskCostGradient(soc::SocAbstraction& soci,arr& dCdq,int t){
  uint n=soci.qDim(),m=soci.nTasks();
  uint i;
  dCdq.resize(n);
  dCdq.setZero();
  arr phiHatQ,J,tJ,x,v,Jqd;
  double prec,precv;
  for(i=0;i<m;i++) if(soci.isConditioned(i,t)){
    soci.getPhi      (phiHatQ,i);
    soci.getTarget   (x,i,t);
    soci.getPrecision(prec,i,t);
    soci.getJtJ      (J,tJ,i);
    dCdq += tJ * ((phiHatQ-x)*(2.*prec));
    if(soci.dynamic){
      soci.getJqd       (Jqd,i);
      soci.getTargetV   (v,i,t);
      soci.getPrecisionV(precv,i,t);
      dCdq += tJ * ((v - Jqd)*(2.*precv));
    }
  }
}

double soc::SocAbstraction::computeTotalCost(const arr& q,bool plot){
  //if(dynamic){ return getTotalDynamicCost(*this,q,os); }
  uint t,T=nTime();
  CHECK(q.nd==2 && q.d0==T && q.d1==qDim(),"q has wrong dimension: "<<q.getDim());
  double taskCost=0., ctrlCost=0.;
  arr W,H,M,F;
  double tau,tau_1=1.,tau_2=1.;
  if(!dynamic){
    getW(W);
  }else{
    getH(H);
    tau=getTau();
    tau_1 = 1./tau;
    tau_2 = tau_1*tau_1;
  }
  arr taskC(T);  taskC.setZero();
  arr ctrlC(T);   ctrlC.setZero();
  for(t=0;t<T;t++){
    if(!dynamic || !t) setq(q[t]);
    else setqv(q[t],tau_1*(q[t]-q[t-1]));
    taskCost += taskC(t) = getTaskCost(*this,t);  //cout <<"cost = " <<cost_t <<endl;
    if(!dynamic){
      if(t>0) ctrlCost += ctrlC(t) =sqrDistance(W,q[t-1],q[t]);
    }else{
      getMF(M,F);
      if(t<T-1 && t>0)
        ctrlCost += sqrDistance(H,tau_2*M*(q[t+1]+q[t-1]-2.*q[t]),F);
      if(t==0)
        ctrlCost += sqrDistance(H,tau_2*M*(q[t+1]-q[t]),F);
    }
  }
  if(plot){
    std::ofstream fil;
    MT::open(fil,"z.trana");
    for(t=0;t<T;t++){
      fil
          <<"time " <<t
          <<"  ctrlC " <<ctrlC(t)
          <<"  taskC " <<taskC(t)
          <<"  totC "  <<ctrlC(t)+taskC(t)
          <<"  q " <<q[t]
          <<endl;
    }
    gnuplot("plot 'z.trana' us 0:4 title 'ctrl costs', 'z.trana' us 0:6 title 'task costs', 'z.trana' us 0:8 title 'tot costs'");
  }
  if(os) *os
        <<"  task-cost " <<taskCost
        <<"  control-cost " <<ctrlCost
        <<"  total-cost " <<taskCost+ctrlCost <<endl;
  return taskCost+ctrlCost;
}

void getTotalDynamicCostGradient(soc::SocAbstraction& soci,const arr& q,arr& dCdq);

void soc::SocAbstraction::computeTotalCostGradient(arr& dCdq,const arr& q){
  if(dynamic){ getTotalDynamicCostGradient(*this,q,dCdq); return; }
  uint t,T=nTime();
  CHECK(q.nd==2 && q.d0==T && q.d1==qDim(),"");
  dCdq.resizeAs(q);
  dCdq.setZero();
  double length=.0, cost1=0., cost2=0.;
  double cost_t;
  arr W;
  getW(W);
  for(t=0;t<T;t++){
    setq(q[t]);
    cost1 += cost_t = getTaskCost(*this,t);  //cout <<"cost = " <<cost_t <<endl;
    getTaskCostGradient(*this,dCdq[t](),t);
    if(t>0) cost2 += sqrDistance(W,q[t-1],q[t]);
    if(t>0) length += metricDistance(W,q[t-1],q[t]);
    if(t>0)   dCdq[t]() += 2.*W*(q[t]-q[t-1]);
    if(t+1<T) dCdq[t]() -= 2.*W*(q[t+1]-q[t]);
  }
}

double getTotalDynamicCost_obsolete(soc::SocAbstraction& soci,const arr& q,std::ostream* os){
  uint t,T=soci.nTime(),i,m=soci.nTasks();
  CHECK(q.nd==2 && q.d0==T && q.d1==soci.qDim(),"q has wrong dimension: "<<q.getDim());
  double taskCost=0., controlCost=0.;
  arr H;
  soci.getH(H);
  double tau,tau_1,tau_2;
  tau=soci.getTau();
  tau_1 = 1./tau;
  tau_2 = tau_1*tau_1;
  arr phiHatQ,x,v,J,tJ,M,F;
  double prec,precv;
  for(t=0;t<T;t++){
    if(!t) soci.setq(q[t]);
    else   soci.setqv(q[t],tau_1*(q[t]-q[t-1]));

    for(i=0;i<m;i++) if(soci.isConditioned(i,t)){
      soci.getPhi      (phiHatQ,i);
      soci.getTarget   (x,i,t);
      soci.getPrecision(prec,i,t);
      taskCost += prec * sqrDistance(phiHatQ,x);

      soci.getJtJ      (J,tJ,i);
      soci.getTargetV   (v,i,t);
      soci.getPrecisionV(precv,i,t);
      if(t>0)
        taskCost += precv * sqrDistance(J*(tau_1*(q[t]-q[t-1])),v);
    }

    soci.getMF(M,F);
    if(t<T-1 && t>0)
      controlCost += sqrDistance(H,tau_2*M*(q[t+1]+q[t-1]-2.*q[t]),F);
    if(t==0)
      controlCost += sqrDistance(H,tau_2*M*(q[t+1]-q[t]),F);
  }
#if 1
  if(os) *os
    <<"  task-cost " <<taskCost
    <<"  control-cost " <<controlCost
    <<"  total-cost " <<taskCost+controlCost <<endl;
#else
  if(os) *os
    <<' ' <<taskCost
    <<' ' <<controlCost
    <<' ' <<taskCost+controlCost <<endl;
#endif
  return taskCost+controlCost;
}

void getTotalDynamicCostGradient(soc::SocAbstraction& soci,const arr& q,arr& dCdq){
  uint t,T=soci.nTime(),i,m=soci.nTasks();
  CHECK(q.nd==2 && q.d0==T && q.d1==soci.qDim(),"");
  dCdq.resizeAs(q);
  dCdq.setZero();
  arr H;
  soci.getH(H);
  double tau,tau_1,tau_2;
  tau=soci.getTau();
  tau_1 = 1./tau;
  tau_2 = tau_1*tau_1;
  arr phiHatQ,x,v,J,tJ,M,Mt,F;
  double prec,precv;
  for(t=0;t<T;t++){
    if(!t) soci.setq(q[t]);
    else   soci.setqv(q[t],tau_1*(q[t]-q[t-1]));

    for(i=0;i<m;i++) if(soci.isConditioned(i,t)){
      soci.getPhi      (phiHatQ,i);
      soci.getTarget   (x,i,t);
      soci.getPrecision(prec,i,t);
      soci.getJtJ      (J,tJ,i);
      dCdq[t]() += (2.*prec)*(tJ * (phiHatQ-x));

      soci.getTargetV   (v,i,t);
      soci.getPrecisionV(precv,i,t);
      if(t>0){
        dCdq[t  ]() += (2.*precv*tau_1)*(tJ * (J*(tau_1*(q[t]-q[t-1])) - v));
        dCdq[t-1]() -= (2.*precv*tau_1)*(tJ * (J*(tau_1*(q[t]-q[t-1])) - v));
      }
    }

    soci.getMF(M,F);
    transpose(Mt,M);
    if(t<T-1 && t>0){
      dCdq[t  ]() -= (4.*tau_2)*Mt*H*(tau_2*M*(q[t+1]+q[t-1]-2.*q[t])-F);
      dCdq[t+1]() += (2.*tau_2)*Mt*H*(tau_2*M*(q[t+1]+q[t-1]-2.*q[t])-F);
      dCdq[t-1]() += (2.*tau_2)*Mt*H*(tau_2*M*(q[t+1]+q[t-1]-2.*q[t])-F);
    }
    if(t==0){
      dCdq[t  ]() -= (2.*tau_2)*Mt*H*(tau_2*M*(q[t+1]-q[t])-F);
      dCdq[t+1]() += (2.*tau_2)*Mt*H*(tau_2*M*(q[t+1]-q[t])-F);
    }
  }
}

//! play the trajectory using OpenGL
void soc::SocAbstraction::displayUpdate(const char *text,bool wait){
  gl->text.clr() <<text;
  if(!wait) gl->update(); else gl->watch();
}

void soc::SocAbstraction::displayTrajectory(const arr& q,int display,const char *tag){
  uint k,t,T=nTime();
  if(!gl || display<=0) return;
  uint num;
  if(display==1) num=T; else num=display;
  for(k=0;k<(uint)num;k++){
    t = k*(T-1)/(num-1);
    setq(q[t]);
    gl->text.clr() <<tag <<" (time " <<std::setw(3) <<t <<'/' <<T <<')';
    gl->update();
    //gl->timedupdate(getTau()*(T-1)/(display-1));
  }
  if(display==1) gl->watch();
}

//! computes separate costs for each ctrl variable
double soc::SocAbstraction::analyzeDynamicTrajectory(const arr& q){
  uint t,T=nTime(),i,m=nTasks();
  CHECK(q.nd==2 && q.d0==T && q.d1==qDim(),"");
  double taskCost=0., controlCost=0.;
  arr H;
  getH(H);
  double tau,tau_1,tau_2;
  tau=getTau();
  tau_1 = 1./tau;
  tau_2 = tau_1*tau_1;
  arr phiHatQ,x,v,J,tJ,M,F;
  double dx,dv,prec,precv;

  arr u;
  arr taskC(T,m);  taskC.setZero();
  arr taskDx(T,m); taskDx.setZero();
  arr taskDv(T,m); taskDv.setZero();
  arr ctrlC(T);   ctrlC.setZero();
  for(t=0;t<T;t++){
    if(!t) setq(q[t]);
    else   setqv(q[t],tau_1*(q[t]-q[t-1]));

    for(i=0;i<m;i++) if(isConditioned(i,t)){
      getPhi      (phiHatQ,i);
      getTarget   (x,i,t);
      getPrecision(prec,i,t);
      getJtJ      (J,tJ,i);
      dx = sqrDistance(phiHatQ,x);
      taskCost += prec*dx;
      taskC(t,i) =prec*dx;
      taskDx(t,i)=sqrt(dx);

      if(t>0){
        getTargetV   (v,i,t);
        getPrecisionV(precv,i,t);
        dv = sqrDistance(J*(tau_1*(q[t]-q[t-1])),v);
        taskCost += precv*dv;
        taskC(t,i)+=precv*dv;
        taskDv(t,i)=sqrt(dv);
      }
    }

    getMF(M,F);
    if(t==0)         u = tau_2*M*(q[t+1]-q[t]) - F;
    if(t<T-1 && t>0) u = tau_2*M*(q[t+1]+q[t-1]-2.*q[t]) - F;
    if(t==T-1)       u.setZero();
    ctrlC(t) = scalarProduct(H,u,u);
    controlCost += ctrlC(t);
    //cout <<"analyzeDynamicTrajectory: time "<<t<<" u=" <<u <<" q=" <<q[t] <<(t>0?tau_1*(q[t]-q[t-1]):q[t]) <<endl;
  }
  /*
  std::ofstream fil;
  MT::open(fil,"z.trana");
  for(t=0;t<T;t++){
    fil
      <<"time " <<t*tau
      <<"  ctrlC " <<ctrlC(t)
      <<"  taskC " <<taskC[t]
      <<"  taskDx " <<taskDx[t]
      <<"  taskDv " <<taskDv[t]
      <<"  q " <<q[t]
      <<endl;
  }
  gnuplot("plot 'z.trana' us 0:4 title 'ctrlC', 'z.trana' us 0:7 title 'taskC', 'z.trana' us 0:19 title 'q'");
  */
  plotClear();
  plotFunctions(q);
  plot(false);
  if(os) *os
    <<"  task-cost " <<taskCost
    <<"  control-cost " <<controlCost
    <<"  total-cost " <<taskCost+controlCost <<endl;
  return taskCost+controlCost;
}
