/*  Copyright 2009 Marc Toussaint
    email: mtoussai@cs.tu-berlin.de

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a COPYING file of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/> */

#include "soc.h"
#include "util.h"

/*! \brief iterated LQG (linear quadratic Gaussian) applied to the
    kinematic trajectory optimization case. This means that we assume
    A=B=1 (simple additive control) with control cost H=W
    (corresponding directly to the q-space metric). The quadratic cost
    terms are computed from the tast constraints */
void soc::iLQG_kinematic(SocAbstraction& soci,
                           arr& q,uint iterations,
                           double convergenceRate,
                           uint display){
  uint t,T=soci.nTime(),n=soci.qDim();

  arr V(T,n,n),v(T,n);
  arr R(T,n,n),r(T,n);
  arr HVinv(T,n,n),VHVinv;
  //we assume A=B=\id for now
  arr H;

  soci.getW(H);

  //first step: initialize with some trajectory
  if(q.N){
    CHECK(q.nd==2 && q.d0==T && q.d1==n,"");
  }else{
    q.resize(T,n);
    bayesianIKTrajectory(soci,q);
  }
  
  MT::timerStart();
  if(soci.os){
#ifdef NIKOLAY
    *soci.os <<std::setw(3) <<-1 << " " <<MT::timerRead(false);
#else
    *soci.os <<std::setw(3) <<-1 <<"  time " <<MT::timerRead(false);
#endif
    soci.computeTotalCost(q);
  }
  if(soci.gl){
    soci.displayTrajectory(q,display,STRING("iLQG optimization -- iteration "<<-1));
  }

  for(uint k=0;k<iterations;k++){
    //linearize around current trajectory
    for(t=0;t<T;t++){
      soci.setq(q[t]);
      getQuadraticTaskCost(soci,t,R[t](),r[t](),q[t]);
    }

    //bwd Ricatti equations
    V[T-1]() = R[T-1];
    v[T-1]() = r[T-1];
    for(t=T-1;t--;){
      inverse_SymPosDef(HVinv[t+1](), H+V[t+1]);
      VHVinv = V[t+1]*HVinv[t+1];
      V[t]() = R[t] + V[t+1] - VHVinv*V[t+1];
      v[t]() = r[t] + v[t+1] - VHVinv*v[t+1];
    }

    //fwd with optimal control
    soci.getq0(q[0]());
    for(t=1;t<T;t++){
      if(convergenceRate==1.){
        q[t]() = q[t-1] - HVinv[t]*(.5*v[t] + V[t]*q[t-1]);
      }else{
        q[t]() = (1.-convergenceRate)*q[t]
               + convergenceRate*(q[t-1] - HVinv[t]*(.5*v[t] + V[t]*q[t-1]));
      }
    }

    if(soci.os){
#ifdef NIKOLAY
      *soci.os <<std::setw(3) << k << " " << MT::timerRead(false);
#else
      *soci.os <<std::setw(3) << k <<"  time " <<MT::timerRead(false);
#endif
      
      soci.computeTotalCost(q);
    }
    if(soci.gl){
      soci.displayTrajectory(q,display,STRING("gradient optimization -- iteration "<<k));
    }
  }
}

//===========================================================================

/*! \brief iterated LQG (linear quadratic Gaussian) for the general
    Stochastic Optimal Control case */
void soc::iLQG_general(SocAbstraction& soci,
		       arr& q,uint iterations,
		       double convergenceRate,
		       uint display){
  CHECK(soci.dynamic,"assumed dynamic SOC abstraction");
  uint t,T=soci.nTime(),n=soci.qDim(),n2;

  n2=2*n;
  arr _q(T,n2);
  arr V(T,n2,n2),v(T,n2);
  arr R(T,n2,n2),r(T,n2);
  arr A(T,n2,n2),a(T,n2),B(T,n2,n);
  arr At,Bt,VA,HVinv,BIG,u; //helpers
  arr H;

  soci.getH(H);
  double tau=soci.getTau();

  //first step: initialize with some trajectory
  if(q.N){
    CHECK(q.nd==2 && q.d0==T && q.d1==n,"");
  }else{
    q.resize(T,n);
    //bayesianIKTrajectory(soci,q);
    soci.passiveDynamicTrajectory(q);
    //soci.getq0(q[0]()); for(t=1;t<T;t++) q[t]()=q[0];
  }
  if(soci.os){
    *soci.os <<std::setw(3) <<-1 <<"  time " <<MT::timerRead(false);
    soci.analyzeDynamicTrajectory(q);
    //soci.computeTotalCost(q);
  }
  if(soci.gl){
    soci.displayTrajectory(q,display,STRING("iLQG optimization -- iteration "<<-1));
  }

  MT::timerStart();
  //cout <<"**Ricatti" <<endl;
  //cout <<"H=" <<H <<endl;

  for(uint k=0;k<iterations;k++){
    //get velocity profile
    getPhaseTrajectory(_q,q,tau);

    //linearize around current trajectory
    for(t=0;t<T;t++){
      soci.setqv(_q[t]);
      soci.getCosts  (R[t](),r[t](),t,q[t]);
      soci.getProcess(A[t](),a[t](),B[t]());
      //cout <<"t=" <<t <<" A=" <<A[t] <<" a=" <<a[t] <<" B=" <<B[t] <<endl;
      //cout <<"t=" <<t <<" R=" <<R[t] <<" r=" <<r[t] <<endl;
    }

    //bwd Ricatti equations
    V[T-1]() = R[T-1];
    v[T-1]() = r[T-1];
    for(t=T-1;t--;){
      transpose(At,A[t]);
      transpose(Bt,B[t]);
      innerProduct(VA,V[t+1],A[t]);
      inverse_SymPosDef(HVinv, H+Bt*V[t+1]*B[t]);
      BIG = (~VA)*B[t]*HVinv*Bt;
      V[t]() = R[t] + (At-BIG)*VA;
      v[t]() = r[t] + (At-BIG)*(v[t+1]+2.*V[t+1]*a[t]);
      //cout <<"t=" <<t <<" V=" <<V[t] <<" v=" <<v[t] <<endl;
    }

    //MT::wait();

    double ctrlC=0;
    //fwd with optimal control
    //we assume that _q[0] is fine and fixed!    (from getPhaseTrajectory)
    for(t=0;t<T-1;t++){
      soci.setqv(_q[t]);
      soci.getProcess(A[t](),a[t](),B[t]());

      transpose(Bt,B[t]);
      inverse_SymPosDef(HVinv, H+Bt*V[t+1]*B[t]);
      u = - HVinv*Bt*(.5*v[t+1] + V[t+1]*(A[t]*_q[t]+a[t]));
      ctrlC += scalarProduct(H,u,u);
      //cout <<"ilqg time "<<t<<" u=" <<u <<" q=" <<_q[t] <<endl;
#if 0
      _q[t+1]() = A[t]*_q[t] + a[t] + B[t]*u;
#else
      if(convergenceRate==1.){
        _q[t+1]() = A[t]*_q[t] + a[t] + B[t]*u;
      }else{
        _q.reshape(T,2,n);
        _q.subDim(t+1,1)
          = (1.-convergenceRate)*_q.subDim(t+1,1)
          + convergenceRate*(_q.subDim(t,1) + a[t].sub(n,-1) + B[t].sub(n,-1,0,-1)*u);
        _q.subDim(t+1,0) = _q.subDim(t,0) + tau*_q.subDim(t+1,1);
        _q.reshape(T,2*n);
      }
#endif
    }

    cout <<"iLQG says ctrlC=" <<ctrlC <<endl;

    _q.reshape(T,2,n);
#if 0
    if(convergenceRate==1.){
      for(t=0;t<T;t++) q[t]=_q.subDim(t,0);
    }else{
      for(t=0;t<T;t++) q[t]=(1.-convergenceRate)*q[t] + convergenceRate*_q.subDim(t,0);
    }
#else
    for(t=0;t<T;t++) q[t]=_q.subDim(t,0);
#endif

    if(soci.os){
      *soci.os <<std::setw(3) <<k <<"  time " <<MT::timerRead(false);
      soci.analyzeDynamicTrajectory(q);
      //soci.computeTotalCost(q);
    }
    if(soci.gl){
      soci.displayTrajectory(q,display,STRING("gradient optimization -- iteration "<<k));
    }
  }
}
